#include <torch/extension.h>
#include <vector>
#include <cmath>

#define TILE_DIM 16

// CUDA kernel for element-wise LSTM computations (optimized for load balancing)
__global__ void lstm_elementwise_forward(
    const float* __restrict__ gates,
    const float* __restrict__ prev_c,
    float* __restrict__ h,
    float* __restrict__ c,
    int batch_size,
    int hidden_size
) {
    int idx = blockIdx.x * blockDim.x + threadIdx.x;
    int total_elements = batch_size * hidden_size;
    int stride = blockDim.x * gridDim.x;

    for (int i = idx; i < total_elements; i += stride) {
        int b = i / hidden_size;
        int n = i % hidden_size;
        int gate_idx = b * hidden_size * 4 + n;

        float i_gate = gates[gate_idx];
        float f_gate = gates[gate_idx + hidden_size];
        float g_gate = gates[gate_idx + 2 * hidden_size];
        float o_gate = gates[gate_idx + 3 * hidden_size];

        // Apply activations
        i_gate = 1.0f / (1.0f + expf(-i_gate));
        f_gate = 1.0f / (1.0f + expf(-f_gate));
        g_gate = tanhf(g_gate);
        o_gate = 1.0f / (1.0f + expf(-o_gate));

        float c_new = f_gate * prev_c[i] + i_gate * g_gate;
        float h_new = o_gate * tanhf(c_new);

        c[i] = c_new;
        h[i] = h_new;
    }
}

// Optimized linear forward kernel using tiling, loop unrolling, and load balancing
__global__ void tiled_linear_forward_kernel(
    const float* __restrict__ input,
    const float* __restrict__ weight,
    const float* __restrict__ bias,
    float* __restrict__ output,
    int batch_size,
    int input_dim,
    int output_dim
) {
    // Identify the row and column of the output element
    int row = blockIdx.x * TILE_DIM + threadIdx.y;
    int col = blockIdx.y * TILE_DIM + threadIdx.x;
    float sum = 0.f;

    // Shared memory for tiles from input and weight
    __shared__ float shared_A[TILE_DIM][TILE_DIM];
    __shared__ float shared_B[TILE_DIM][TILE_DIM];

    // Number of tiles to cover input_dim
    int numTiles = (input_dim + TILE_DIM - 1) / TILE_DIM;
    for (int t = 0; t < numTiles; t++) {
        int colA = t * TILE_DIM + threadIdx.x; // Column index for input
        int rowB = t * TILE_DIM + threadIdx.y; // Row index for weight tile

        // Load tile from input
        if (row < batch_size && colA < input_dim)
            shared_A[threadIdx.y][threadIdx.x] = input[row * input_dim + colA];
        else
            shared_A[threadIdx.y][threadIdx.x] = 0.f;

        // Load tile from weight
        if (col < output_dim && rowB < input_dim)
            shared_B[threadIdx.y][threadIdx.x] = weight[col * input_dim + rowB];
        else
            shared_B[threadIdx.y][threadIdx.x] = 0.f;

        __syncthreads();

        // Utilize loop unrolling for better performance
        #pragma unroll
        for (int k = 0; k < TILE_DIM; k++) {
            sum += shared_A[threadIdx.y][k] * shared_B[k][threadIdx.x];
        }

        __syncthreads();
    }

    if (row < batch_size && col < output_dim) {
        if (bias != nullptr)
            sum += bias[col];
        output[row * output_dim + col] = sum;
    }
}

// LSTM forward function using the lstm_elementwise_forward kernel
torch::Tensor lstm_forward_cuda(
    torch::Tensor input,
    torch::Tensor w_ih,
    torch::Tensor w_hh,
    torch::Tensor b_ih,
    torch::Tensor b_hh,
    torch::Tensor h0,
    torch::Tensor c0
) {
    int batch_size = input.size(0);
    int seq_len = input.size(1);
    int input_size = input.size(2);
    int hidden_size = h0.size(1);

    torch::Tensor h = h0.clone();
    torch::Tensor c = c0.clone();

    std::vector<torch::Tensor> outputs;

    int threads = 256;
    int blocks = std::min(65535, (batch_size * hidden_size + threads - 1) / threads);

    for (int t = 0; t < seq_len; t++) {
        torch::Tensor xt = input.select(1, t);

        torch::Tensor gates = torch::addmm(b_ih, xt, w_ih.t());
        gates = torch::addmm(gates, h, w_hh.t());
        gates += b_hh;

        const float* gates_ptr = gates.data_ptr<float>();
        const float* prev_c_ptr = c.data_ptr<float>();
        float* h_ptr = h.data_ptr<float>();
        float* c_ptr = c.data_ptr<float>();

        lstm_elementwise_forward<<<blocks, threads>>>(
            gates_ptr,
            prev_c_ptr,
            h_ptr,
            c_ptr,
            batch_size,
            hidden_size
        );

        outputs.push_back(h.unsqueeze(1));
    }

    torch::Tensor output = torch::cat(outputs, 1);
    return output;
}

// Linear forward function using a tiled and unrolled GEMM kernel
torch::Tensor linear_forward_cuda(
    torch::Tensor input,
    torch::Tensor weight,
    torch::Tensor bias
) {
    int batch_size = input.size(0);
    int input_dim = input.size(1);
    int output_dim = weight.size(0);

    auto options = torch::TensorOptions().dtype(input.dtype()).device(input.device());
    auto output = torch::empty({batch_size, output_dim}, options);

    dim3 block(TILE_DIM, TILE_DIM);
    dim3 grid((batch_size + TILE_DIM - 1) / TILE_DIM, (output_dim + TILE_DIM - 1) / TILE_DIM);

    tiled_linear_forward_kernel<<<grid, block>>>(
        input.data_ptr<float>(),
        weight.data_ptr<float>(),
        bias.defined() ? bias.data_ptr<float>() : nullptr,
        output.data_ptr<float>(),
        batch_size,
        input_dim,
        output_dim
    );

    return output;
}

// Main forward function orchestrating LSTM layers and the final linear layer
torch::Tensor forward(
    torch::Tensor x,
    std::vector<torch::Tensor> lstm_weights_ih,
    std::vector<torch::Tensor> lstm_weights_hh,
    std::vector<torch::Tensor> lstm_biases_ih,
    std::vector<torch::Tensor> lstm_biases_hh,
    torch::Tensor fc_weight,
    torch::Tensor fc_bias,
    torch::Tensor h0,
    torch::Tensor c0,
    bool is_training
) {
    h0 = h0.to(x.device());
    c0 = c0.to(x.device());
    
    torch::Tensor out = x;
    int num_layers = lstm_weights_ih.size();

    for (int i = 0; i < num_layers; i++) {
        auto w_ih = lstm_weights_ih[i].to(x.device());
        auto w_hh = lstm_weights_hh[i].to(x.device());
        auto b_ih = lstm_biases_ih[i].to(x.device());
        auto b_hh = lstm_biases_hh[i].to(x.device());

        torch::Tensor h_i = h0.narrow(0, i, 1).squeeze(0);
        torch::Tensor c_i = c0.narrow(0, i, 1).squeeze(0);
        
        out = lstm_forward_cuda(out, w_ih, w_hh, b_ih, b_hh, h_i, c_i);
    }

    // Take the last timestep and apply the final linear layer
    out = out.select(1, -1);
    out = linear_forward_cuda(out, fc_weight, fc_bias);
    
    return out;
}

PYBIND11_MODULE(TORCH_EXTENSION_NAME, m) {
    m.def("forward", &forward, "Module forward (CUDA) with load balancing and optimization");
}
